@@ -1,14 +1,16 @@
 import sys
-from .shared import *
-from .analytics_helper import *
-from .model_index import *
-from .opc_ua import *
 
 if sys.version_info[:2] >= (3, 8):
     # TODO: Import directly (no need for conditional) when `python_requires = >= 3.9`
-    from importlib.metadata import PackageNotFoundError, version  # pragma: no cover
+    from importlib.metadata import (  # pragma: no cover
+        PackageNotFoundError,
+        version,
+    )
 else:
-    from importlib_metadata import PackageNotFoundError, version  # pragma: no cover
+    from importlib_metadata import (  # pragma: no cover
+        PackageNotFoundError,
+        version,
+    )
 
 try:
     # Change here if project is renamed and does not equal the package name
