@@ -1,6 +1,7 @@
 import json
+from typing import Dict, List, Union
+
 from pydantic import validate_call
-from typing import List, Dict, Union
 
 from ..idwh import IDWH
 
@@ -31,4 +32,6 @@ class Solcast:
         )
 
         query = "EXEC dwetl.UpsertSolcastForecastData ?, ?"
-        return self.dwh.execute(query, solcast_forecast_data_json, forecast_type_key)
+        return self.dwh.execute(
+            query, solcast_forecast_data_json, forecast_type_key
+        )
