@@ -1,19 +1,21 @@
+import asyncio
+import copy
 import json
 import logging
-import copy
+from asyncio import Semaphore
+from datetime import date, datetime, timedelta
+from typing import Any, Callable, Dict, List, Optional, Union
+
+import aiohttp
+import nest_asyncio
 import pandas as pd
 import requests
-from datetime import date, datetime, timedelta
-from typing import Dict, List, Any, Union, Optional, Callable
-from pydantic import BaseModel, AnyUrl
-from pydantic_core import Url
-from pyprediktormapclient.shared import request_from_api
-from requests import HTTPError
-import asyncio
-import aiohttp
 from aiohttp import ClientSession
-from asyncio import Semaphore
-import nest_asyncio
+from pydantic import AnyUrl, BaseModel
+from pydantic_core import Url
+from requests import HTTPError
+
+from pyprediktormapclient.shared import request_from_api
 
 nest_asyncio.apply()
 
@@ -31,49 +33,58 @@ class Variables(BaseModel):
             Namespace: int - Namespace on the signal, e.g. 2.
             IdType: int - IdTypes described in https://reference.opcfoundation.org/v104/Core/docs/Part3/8.2.3/.
     """
+
     Id: str
     Namespace: int
     IdType: int
 
+
 class SubValue(BaseModel):
     """Helper class to parse all values api's.
 
-        Variables:
-            Type: int - Type of variable, e.g. 12. string, 11. float, etc. list of types described in https://reference.opcfoundation.org/Core/Part6/v104/5.1.2/ 
-            Body: Union[str, float, int, bool] - The value of the varible, should match type.
+    Variables:
+        Type: int - Type of variable, e.g. 12. string, 11. float, etc. list of types described in https://reference.opcfoundation.org/Core/Part6/v104/5.1.2/
+        Body: Union[str, float, int, bool] - The value of the varible, should match type.
     """
+
     Type: int
     Body: Union[str, float, int, bool]
 
+
 class HistoryValue(BaseModel):
     """Helper class to parse all values api's.
 
-        Variables:
-            Value: SubValue - Containing Type and Body (value) of the variable. Described in SubValue class.
+    Variables:
+        Value: SubValue - Containing Type and Body (value) of the variable. Described in SubValue class.
     """
+
     Value: SubValue
 
+
 class StatsCode(BaseModel):
     """Helper class to parse all values api's.
 
-        Variables:
-            Code: Optional[int] - Status code, described in https://reference.opcfoundation.org/v104/Core/docs/Part8/A.4.3/
-            Symbol: Optional[str] - String value for status code, described in https://reference.opcfoundation.org/v104/Core/docs/Part8/A.4.3/
+    Variables:
+        Code: Optional[int] - Status code, described in https://reference.opcfoundation.org/v104/Core/docs/Part8/A.4.3/
+        Symbol: Optional[str] - String value for status code, described in https://reference.opcfoundation.org/v104/Core/docs/Part8/A.4.3/
     """
+
     Code: Optional[int] = None
     Symbol: Optional[str] = None
 
+
 class Value(BaseModel):
     """Helper class to parse all values api's.
 
-        Variables:
-            Value: SubValue - Containing Type and Body (value) of the variable. Described in SubValue class.
-            SourceTimestamp: str - Timestamp of the source, e.g. when coming from an API the timestamp returned from the API for the varaible is the sourcetimestamp.
-            SourcePicoseconds: Optional[int] - Picoseconds for the timestamp of the source if there is a need for a finer granularity, e.g. if samples are sampled in picosecond level or more precision is needed.
-            ServerTimestamp: Optional[str] - Timestamp for the server, normally this is assigned by the server.
-            ServerPicoseconds: Optional[int] - Picoseconds for the timestamp on the server, normally this is assigned by the server.
-            StatusCode: StatusCode - Status code, described in https://reference.opcfoundation.org/v104/Core/docs/Part8/A.4.3/
+    Variables:
+        Value: SubValue - Containing Type and Body (value) of the variable. Described in SubValue class.
+        SourceTimestamp: str - Timestamp of the source, e.g. when coming from an API the timestamp returned from the API for the varaible is the sourcetimestamp.
+        SourcePicoseconds: Optional[int] - Picoseconds for the timestamp of the source if there is a need for a finer granularity, e.g. if samples are sampled in picosecond level or more precision is needed.
+        ServerTimestamp: Optional[str] - Timestamp for the server, normally this is assigned by the server.
+        ServerPicoseconds: Optional[int] - Picoseconds for the timestamp on the server, normally this is assigned by the server.
+        StatusCode: StatusCode - Status code, described in https://reference.opcfoundation.org/v104/Core/docs/Part8/A.4.3/
     """
+
     Value: SubValue
     SourceTimestamp: datetime
     SourcePicoseconds: Optional[int] = None
@@ -81,45 +92,54 @@ class Value(BaseModel):
     ServerPicoseconds: Optional[int] = None
     StatusCode: Optional[StatsCode] = None
 
+
 class WriteVariables(BaseModel):
     """Helper class for write values api.
 
-        Variables:
-                NodeId: Variables - The complete node'id for the variable
-                Value: Value - The value to update for the node'id.
+    Variables:
+            NodeId: Variables - The complete node'id for the variable
+            Value: Value - The value to update for the node'id.
     """
+
     NodeId: Variables
     Value: Value
 
+
 class WriteHistoricalVariables(BaseModel):
     """Helper class for write historical values api.
 
-        Variables:
-            NodeId (str): The complete node'id for the variable
-            PerformInsertReplace (int): Historical insertion method 1. Insert, 2. Replace 3. Update, 4. Remove
-            UpdateValues (list): List of values to update for the node'id. Time must be in descending order.
+    Variables:
+        NodeId (str): The complete node'id for the variable
+        PerformInsertReplace (int): Historical insertion method 1. Insert, 2. Replace 3. Update, 4. Remove
+        UpdateValues (list): List of values to update for the node'id. Time must be in descending order.
     """
+
     NodeId: Variables
     PerformInsertReplace: int
     UpdateValues: List[Value]
 
+
 class WriteVariablesResponse(BaseModel):
     """Helper class for write historical values api.
 
-        Variables:
-            SymbolCodes: List[StatusCode] - A list of class StatusCode, described in StatusCode class.
+    Variables:
+        SymbolCodes: List[StatusCode] - A list of class StatusCode, described in StatusCode class.
     """
+
     SymbolCodes: List[StatsCode]
 
-class WriteReturn(BaseModel):
-    """Helper class to collect API output with API input to see successfull writes for nodes.
 
-        Variables:
-            Id: str - The Id of the signal
-            Value: str - The written value of the signal
-            TimeStamp: str - THe SourceTimestamp of the written signal
-            Success: bool - Success flag for the write operation
+class WriteReturn(BaseModel):
+    """Helper class to collect API output with API input to see successfull
+    writes for nodes.
+
+    Variables:
+        Id: str - The Id of the signal
+        Value: str - The written value of the signal
+        TimeStamp: str - THe SourceTimestamp of the written signal
+        Success: bool - Success flag for the write operation
     """
+
     Id: str
     Value: str
     TimeStamp: str
@@ -132,11 +152,13 @@ class AsyncIONotebookHelper:
         loop = asyncio.get_event_loop()
         return loop.run_until_complete(coroutine)
 
+
 class Config:
-        arbitrary_types_allowed = True
+    arbitrary_types_allowed = True
+
 
 class OPC_UA:
-    """Helper functions to access the OPC UA REST Values API server
+    """Helper functions to access the OPC UA REST Values API server.
 
     Args:
         rest_url (str): The complete url of the OPC UA Values REST API. E.g. "http://127.0.0.1:13371/"
@@ -151,12 +173,19 @@ class OPC_UA:
         * Better session handling with aiohttp
         * Make sure that time convertions are with timezone
     """
+
     class Config:
         arbitrary_types_allowed = True
 
-    
-    def __init__(self, rest_url: AnyUrl, opcua_url: AnyUrl, namespaces: List = None, auth_client: object = None, session: requests.Session = None):
-        """Class initializer
+    def __init__(
+        self,
+        rest_url: AnyUrl,
+        opcua_url: AnyUrl,
+        namespaces: List = None,
+        auth_client: object = None,
+        session: requests.Session = None,
+    ):
+        """Class initializer.
 
         Args:
             rest_url (str): The complete url of the OPC UA Values REST API. E.g. "http://127.0.0.1:13371/"
@@ -170,36 +199,45 @@ class OPC_UA:
         self.opcua_url = opcua_url
         self.headers = {
             "Content-Type": "application/json",
-            "Accept": "text/plain"
+            "Accept": "text/plain",
         }
         self.auth_client = auth_client
         self.session = session
         self.helper = AsyncIONotebookHelper()
-        
+
+        if not str(self.opcua_url).startswith("opc.tcp://"):
+            raise ValueError("Invalid OPC UA URL")
+
         if self.auth_client is not None:
             if self.auth_client.token is not None:
-                self.headers["Authorization"] = f"Bearer {self.auth_client.token.session_token}"
-        self.body = {"Connection": {"Url": self.opcua_url, "AuthenticationType": 1}}
+                self.headers["Authorization"] = (
+                    f"Bearer {self.auth_client.token.session_token}"
+                )
+        self.body = {
+            "Connection": {"Url": self.opcua_url, "AuthenticationType": 1}
+        }
         if namespaces:
             self.body["ClientNamespaces"] = namespaces
 
     def json_serial(self, obj):
-        """JSON serializer for objects not serializable by default json code"""
+        """JSON serializer for objects not serializable by default json
+        code."""
 
         if isinstance(obj, (datetime, date)):
             return obj.isoformat()
         elif isinstance(obj, Url):
             return str(obj)
-        raise TypeError (f"Type {type(obj)} not serializable")
+        raise TypeError(f"Type {type(obj)} not serializable")
 
     def check_auth_client(self, content):
-        if (content.get('error').get('code') == 404):
+        if content.get("error").get("code") == 404:
             self.auth_client.request_new_ory_token()
-            self.headers["Authorization"] = f"Bearer {self.auth_client.token.session_token}"
+            self.headers["Authorization"] = (
+                f"Bearer {self.auth_client.token.session_token}"
+            )
         else:
             raise RuntimeError(content.get("ErrorMessage"))
 
-    
     def _get_value_type(self, id: int) -> Dict:
         """Internal function to get the type of a value from the OPC UA return,as documentet at
         https://docs.prediktor.com/docs/opcuavaluesrestapi/datatypes.html#variant
@@ -216,7 +254,7 @@ class OPC_UA:
 
     def _get_variable_list_as_list(self, variable_list: list) -> list:
         """Internal function to convert a list of pydantic Variable models to a
-        list of dicts
+        list of dicts.
 
         Args:
             variable_list (List): List of pydantic models
@@ -226,7 +264,7 @@ class OPC_UA:
         """
         new_vars = []
         for var in variable_list:
-            if hasattr(var, 'model_dump'):
+            if hasattr(var, "model_dump"):
                 # Convert pydantic model to dict
                 new_vars.append(var.model_dump())
             elif isinstance(var, dict):
@@ -237,9 +275,8 @@ class OPC_UA:
 
         return new_vars
 
-    
     def get_values(self, variable_list: List[Variables]) -> List:
-        """Request realtime values from the OPC UA server
+        """Request realtime values from the OPC UA server.
 
         Args:
             variable_list (list): A list of variables you want, containing keys "Id", "Namespace" and "IdType"
@@ -260,20 +297,20 @@ class OPC_UA:
                 extended_timeout=True,
             )
         except HTTPError as e:
-            # print(.get('error').get('code'))
             if self.auth_client is not None:
                 self.check_auth_client(json.loads(e.response.content))
+                content = request_from_api(
+                    rest_url=self.rest_url,
+                    method="POST",
+                    endpoint="values/get",
+                    data=json.dumps([body], default=self.json_serial),
+                    headers=self.headers,
+                    extended_timeout=True,
+                )
             else:
-                raise RuntimeError(f'Error message {e}')
-        finally:
-            content = request_from_api(
-                rest_url=self.rest_url,
-                method="POST",
-                endpoint="values/get",
-                data=json.dumps([body], default=self.json_serial),
-                headers=self.headers,
-                extended_timeout=True,
-            )
+                raise RuntimeError(f"Error in get_values: {str(e)}") from e
+        except Exception as e:
+            raise RuntimeError(f"Error in get_values: {str(e)}") from e
 
         for var in vars:
             # Add default None values
@@ -309,11 +346,12 @@ class OPC_UA:
             # StatusCode is not always present in the answer
             if "StatusCode" in contline:
                 vars[num]["StatusCode"] = contline["StatusCode"].get("Code")
-                vars[num]["StatusSymbol"] = contline["StatusCode"].get("Symbol")
+                vars[num]["StatusSymbol"] = contline["StatusCode"].get(
+                    "Symbol"
+                )
 
         return vars
 
-    
     def _check_content(self, content: Dict[str, Any]) -> None:
         """Check the content returned from the server.
 
@@ -328,25 +366,26 @@ class OPC_UA:
         if not content.get("Success"):
             raise RuntimeError(content.get("ErrorMessage"))
         if "HistoryReadResults" not in content:
-            raise RuntimeError(content.get("ErrorMessage"))
+            raise RuntimeError(
+                "No history read results returned from the server"
+            )
 
-    def _process_df(self, df_result: pd.DataFrame, columns: Dict[str, str]) -> pd.DataFrame:
-        """
-        Process the DataFrame returned from the server.
-        """
+    def _process_df(
+        self, df_result: pd.DataFrame, columns: Dict[str, str]
+    ) -> pd.DataFrame:
+        """Process the DataFrame returned from the server."""
         if "Value.Type" in df_result.columns:
-            df_result["Value.Type"] = df_result["Value.Type"].replace(self.TYPE_DICT)
+            df_result["Value.Type"] = df_result["Value.Type"].replace(
+                self.TYPE_DICT
+            )
 
-        df_result.rename(
-            columns=columns, 
-            errors="raise", 
-            inplace=True
-        )
+        df_result.rename(columns=columns, errors="raise", inplace=True)
 
         return df_result
-    
 
-    async def _make_request(self, endpoint: str, body: dict, max_retries: int, retry_delay: int):
+    async def _make_request(
+        self, endpoint: str, body: dict, max_retries: int, retry_delay: int
+    ):
         for attempt in range(max_retries):
             try:
                 logging.info(f"Attempt {attempt + 1} of {max_retries}")
@@ -355,35 +394,42 @@ class OPC_UA:
                     logging.info(f"Making POST request to {url}")
                     logging.debug(f"Request body: {body}")
                     logging.debug(f"Request headers: {self.headers}")
-                    
-                    async with session.post(url, json=body, headers=self.headers) as response:
-                        logging.info(f"Response received: Status {response.status}")
-                        
+
+                    async with session.post(
+                        url, json=body, headers=self.headers
+                    ) as response:
+                        logging.info(
+                            f"Response received: Status {response.status}"
+                        )
+
                         if response.status >= 400:
                             error_text = await response.text()
-                            logging.error(f"HTTP error {response.status}: {error_text}")
-                            response.raise_for_status()  
-                        
+                            logging.error(
+                                f"HTTP error {response.status}: {error_text}"
+                            )
+                            await response.raise_for_status()
+
                         return await response.json()
 
             except aiohttp.ClientResponseError as e:
-                if e.status == 500:
-                    logging.error(f"Server Error: {e}")
-                    raise  # For 500 errors, we might want to fail fast
                 logging.error(f"ClientResponseError: {e}")
+                if attempt == max_retries - 1:
+                    raise RuntimeError("Max retries reached") from e
             except aiohttp.ClientError as e:
                 logging.error(f"ClientError in POST request: {e}")
             except Exception as e:
                 logging.error(f"Unexpected error in _make_request: {e}")
 
             if attempt < max_retries - 1:
-                wait_time = retry_delay * (2 ** attempt)
-                logging.warning(f"Request failed. Retrying in {wait_time} seconds...")
+                wait_time = retry_delay * (2**attempt)
+                logging.warning(
+                    f"Request failed. Retrying in {wait_time} seconds..."
+                )
                 await asyncio.sleep(wait_time)
-            else:
-                logging.error(f"Max retries reached.")
-                raise RuntimeError('Max retries reached')
-                
+
+        logging.error("Max retries reached.")
+        raise RuntimeError("Max retries reached")
+
     def _process_content(self, content: dict) -> pd.DataFrame:
         self._check_content(content)
         df_list = []
@@ -392,12 +438,12 @@ class OPC_UA:
             for key, value in item["NodeId"].items():
                 df[f"HistoryReadResults.NodeId.{key}"] = value
             df_list.append(df)
-        
+
         if df_list:
             df_result = pd.concat(df_list)
             df_result.reset_index(inplace=True, drop=True)
             return df_result
-    
+
     async def get_historical_values(
         self,
         start_time: datetime,
@@ -409,26 +455,36 @@ class OPC_UA:
         max_data_points: int = 10000,
         max_retries: int = 3,
         retry_delay: int = 5,
-        max_concurrent_requests: int = 30
+        max_concurrent_requests: int = 30,
     ) -> pd.DataFrame:
-        """Generic method to request historical values from the OPC UA server with batching"""
+        """Generic method to request historical values from the OPC UA server
+        with batching."""
         total_time_range_ms = (end_time - start_time).total_seconds() * 1000
         estimated_intervals = total_time_range_ms / max_data_points
 
-        max_variables_per_batch = max(1, int(max_data_points / estimated_intervals))
+        max_variables_per_batch = max(
+            1, int(max_data_points / estimated_intervals)
+        )
         max_time_batches = max(1, int(estimated_intervals / max_data_points))
         time_batch_size_ms = total_time_range_ms / max_time_batches
 
         extended_variables = prepare_variables(variable_list)
-        variable_batches = [extended_variables[i:i + max_variables_per_batch] for i in range(0, len(extended_variables), max_variables_per_batch)]
+        variable_batches = [
+            extended_variables[i : i + max_variables_per_batch]
+            for i in range(0, len(extended_variables), max_variables_per_batch)
+        ]
 
         semaphore = Semaphore(max_concurrent_requests)
 
         async def process_batch(variables, time_batch):
             async with semaphore:
                 batch_start_ms = time_batch * time_batch_size_ms
-                batch_end_ms = min((time_batch + 1) * time_batch_size_ms, total_time_range_ms)
-                batch_start = start_time + timedelta(milliseconds=batch_start_ms)
+                batch_end_ms = min(
+                    (time_batch + 1) * time_batch_size_ms, total_time_range_ms
+                )
+                batch_start = start_time + timedelta(
+                    milliseconds=batch_start_ms
+                )
                 batch_end = start_time + timedelta(milliseconds=batch_end_ms)
 
                 body = {
@@ -436,10 +492,12 @@ class OPC_UA:
                     "StartTime": batch_start.isoformat() + "Z",
                     "EndTime": batch_end.isoformat() + "Z",
                     "ReadValueIds": variables,
-                    **(additional_params or {})
+                    **(additional_params or {}),
                 }
 
-                content = await self._make_request(endpoint, body, max_retries, retry_delay)
+                content = await self._make_request(
+                    endpoint, body, max_retries, retry_delay
+                )
                 return self._process_content(content)
 
         tasks = [
@@ -449,10 +507,14 @@ class OPC_UA:
         ]
 
         results = await asyncio.gather(*tasks)
+        results = [df for df in results if df is not None]
+
+        if not results:
+            return pd.DataFrame()
+
         combined_df = pd.concat(results, ignore_index=True)
         return combined_df
 
-
     async def get_historical_raw_values_asyn(
         self,
         start_time: datetime,
@@ -460,13 +522,16 @@ class OPC_UA:
         variable_list: List[str],
         limit_start_index: Union[int, None] = None,
         limit_num_records: Union[int, None] = None,
-        **kwargs
+        **kwargs,
     ) -> pd.DataFrame:
-        """Request raw historical values from the OPC UA server"""
+        """Request raw historical values from the OPC UA server."""
 
         additional_params = {}
         if limit_start_index is not None and limit_num_records is not None:
-            additional_params["Limit"] = {"StartIndex": limit_start_index, "NumRecords": limit_num_records}
+            additional_params["Limit"] = {
+                "StartIndex": limit_start_index,
+                "NumRecords": limit_num_records,
+            }
 
         combined_df = await self.get_historical_values(
             start_time,
@@ -475,7 +540,7 @@ class OPC_UA:
             "values/historical",
             lambda vars: [{"NodeId": var} for var in vars],
             additional_params,
-            **kwargs
+            **kwargs,
         )
         columns = {
             "Value.Type": "ValueType",
@@ -483,14 +548,15 @@ class OPC_UA:
             "SourceTimestamp": "Timestamp",
             "HistoryReadResults.NodeId.IdType": "IdType",
             "HistoryReadResults.NodeId.Id": "Id",
-            "HistoryReadResults.NodeId.Namespace": "Namespace"
+            "HistoryReadResults.NodeId.Namespace": "Namespace",
         }
         return self._process_df(combined_df, columns)
-    
+
     def get_historical_raw_values(self, *args, **kwargs):
-            result = self.helper.run_coroutine(self.get_historical_raw_values_asyn(*args, **kwargs))
-            return result
-           
+        result = self.helper.run_coroutine(
+            self.get_historical_raw_values_asyn(*args, **kwargs)
+        )
+        return result
 
     async def get_historical_aggregated_values_asyn(
         self,
@@ -499,13 +565,13 @@ class OPC_UA:
         pro_interval: int,
         agg_name: str,
         variable_list: List[str],
-        **kwargs
+        **kwargs,
     ) -> pd.DataFrame:
-        """Request historical aggregated values from the OPC UA server"""
+        """Request historical aggregated values from the OPC UA server."""
 
         additional_params = {
             "ProcessingInterval": pro_interval,
-            "AggregateName": agg_name
+            "AggregateName": agg_name,
         }
 
         combined_df = await self.get_historical_values(
@@ -513,9 +579,11 @@ class OPC_UA:
             end_time,
             variable_list,
             "values/historicalaggregated",
-            lambda vars: [{"NodeId": var, "AggregateName": agg_name} for var in vars],
+            lambda vars: [
+                {"NodeId": var, "AggregateName": agg_name} for var in vars
+            ],
             additional_params,
-            **kwargs
+            **kwargs,
         )
         columns = {
             "Value.Type": "ValueType",
@@ -525,17 +593,18 @@ class OPC_UA:
             "SourceTimestamp": "Timestamp",
             "HistoryReadResults.NodeId.IdType": "IdType",
             "HistoryReadResults.NodeId.Id": "Id",
-            "HistoryReadResults.NodeId.Namespace": "Namespace"
+            "HistoryReadResults.NodeId.Namespace": "Namespace",
         }
         return self._process_df(combined_df, columns)
-                
+
     def get_historical_aggregated_values(self, *args, **kwargs):
-        result = self.helper.run_coroutine(self.get_historical_aggregated_values_asyn(*args, **kwargs))
+        result = self.helper.run_coroutine(
+            self.get_historical_aggregated_values_asyn(*args, **kwargs)
+        )
         return result
-    
-    
+
     def write_values(self, variable_list: List[WriteVariables]) -> List:
-        """Request to write realtime values to the OPC UA server
+        """Request to write realtime values to the OPC UA server.
 
         Args:
             variable_list (list): A list of variables you want to write to with the value, timestamp and quality, containing keys "Id", "Namespace", "Values" and "IdType".
@@ -558,34 +627,41 @@ class OPC_UA:
         except HTTPError as e:
             if self.auth_client is not None:
                 self.check_auth_client(json.loads(e.response.content))
+                content = request_from_api(
+                    rest_url=self.rest_url,
+                    method="POST",
+                    endpoint="values/set",
+                    data=json.dumps([body], default=self.json_serial),
+                    headers=self.headers,
+                    extended_timeout=True,
+                )
             else:
-                raise RuntimeError(f'Error message {e}')
-        finally:
-            content = request_from_api(
-                rest_url=self.rest_url,
-                method="POST",
-                endpoint="values/set",
-                data=json.dumps([body], default=self.json_serial),
-                headers=self.headers,
-                extended_timeout=True,
-            )
+                raise RuntimeError(f"Error in write_values: {str(e)}")
+        except Exception as e:
+            raise RuntimeError(f"Error in write_values: {str(e)}")
+
         # Return if no content from server
         if not isinstance(content, dict):
             return None
         if content.get("Success") is False:
             raise RuntimeError(content.get("ErrorMessage"))
         if content.get("StatusCodes") is None:
-            raise ValueError('No status codes returned, might indicate no values written')
+            raise ValueError(
+                "No status codes returned, might indicate no values written"
+            )
 
         # Use to place successfull write next to each written values as API only returns list. Assumes same index in response as in request.
         for num, row in enumerate(vars):
-            vars[num]["WriteSuccess"]=(lambda x : True if(x == 0) else False)(content['StatusCodes'][num].get("Code"))
+            vars[num]["WriteSuccess"] = (
+                lambda x: True if (x == 0) else False
+            )(content["StatusCodes"][num].get("Code"))
 
         return vars
 
-
-    def write_historical_values(self, variable_list: List[WriteHistoricalVariables]) -> List:
-        """Request to write realtime values to the OPC UA server
+    def write_historical_values(
+        self, variable_list: List[WriteHistoricalVariables]
+    ) -> List:
+        """Request to write realtime values to the OPC UA server.
 
         Args:
             variable_list (list): A list of variables you want, containing keys "Id", "Namespace", "Values" and "IdType". Values must be in descending order of the timestamps.
@@ -594,10 +670,21 @@ class OPC_UA:
         """
         # Check if data is in correct order, if wrong fail.
         for variable in variable_list:
-            if(len(variable.UpdateValues)>1):
-                for num_variable in range(len(variable.UpdateValues) - 1):
-                    if not((variable.UpdateValues[num_variable].SourceTimestamp) < variable.UpdateValues[num_variable+1].SourceTimestamp):
-                        raise ValueError("Time for variables not in correct order.")
+            if len(variable.get("UpdateValues", [])) > 1:
+                for num_variable in range(len(variable["UpdateValues"]) - 1):
+                    if not (
+                        (
+                            variable["UpdateValues"][num_variable][
+                                "SourceTimestamp"
+                            ]
+                        )
+                        < variable["UpdateValues"][num_variable + 1][
+                            "SourceTimestamp"
+                        ]
+                    ):
+                        raise ValueError(
+                            "Time for variables not in correct order."
+                        )
         # Create a new variable list to remove pydantic models
         vars = self._get_variable_list_as_list(variable_list)
         body = copy.deepcopy(self.body)
@@ -614,17 +701,21 @@ class OPC_UA:
         except HTTPError as e:
             if self.auth_client is not None:
                 self.check_auth_client(json.loads(e.response.content))
+                # Retry the request after checking auth
+                content = request_from_api(
+                    rest_url=self.rest_url,
+                    method="POST",
+                    endpoint="values/historicalwrite",
+                    data=json.dumps(body, default=self.json_serial),
+                    headers=self.headers,
+                    extended_timeout=True,
+                )
             else:
-                raise RuntimeError(f'Error message {e}')
-        finally:
-            content = request_from_api(
-                rest_url=self.rest_url,
-                method="POST",
-                endpoint="values/historicalwrite",
-                data=json.dumps(body, default=self.json_serial),
-                headers=self.headers,
-                extended_timeout=True,
-            )
+                raise RuntimeError(
+                    f"Error in write_historical_values: {str(e)}"
+                )
+        except Exception as e:
+            raise RuntimeError(f"Error in write_historical_values: {str(e)}")
         # Return if no content from server
         if not isinstance(content, dict):
             return None
@@ -633,7 +724,9 @@ class OPC_UA:
             raise RuntimeError(content.get("ErrorMessage"))
         # Crash if history report is missing
         if content.get("HistoryUpdateResults") is None:
-            raise ValueError('No status codes returned, might indicate no values written')
+            raise ValueError(
+                "No status codes returned, might indicate no values written"
+            )
 
         # Check if there are per history update error codes returned
         for num_var, variable_row in enumerate(vars):
@@ -642,33 +735,53 @@ class OPC_UA:
                 vars[num_var]["WriteSuccess"] = True
             else:
                 vars[num_var]["WriteSuccess"] = False
-                vars[num_var]["WriteError"] = content["HistoryUpdateResults"][num_var].get("StatusCode")
+                vars[num_var]["WriteError"] = content["HistoryUpdateResults"][
+                    num_var
+                ].get("StatusCode")
 
         return vars
 
     def check_if_ory_session_token_is_valid_refresh(self):
-        """Check if the session token is still valid
-
-        """
+        """Check if the session token is still valid."""
         if self.auth_client.check_if_token_has_expired():
             self.auth_client.refresh_token()
 
 
 TYPE_LIST = [
-    {"id": 0, "type": "Null", "description": "An invalid or unspecified value"},
+    {
+        "id": 0,
+        "type": "Null",
+        "description": "An invalid or unspecified value",
+    },
     {
         "id": 1,
         "type": "Boolean",
         "description": "A boolean logic value (true or false)",
     },
     {"id": 2, "type": "SByte", "description": "An 8 bit signed integer value"},
-    {"id": 3, "type": "Byte", "description": "An 8 bit unsigned integer value"},
+    {
+        "id": 3,
+        "type": "Byte",
+        "description": "An 8 bit unsigned integer value",
+    },
     {"id": 4, "type": "Int16", "description": "A 16 bit signed integer value"},
-    {"id": 5, "type": "UInt16", "description": "A 16 bit unsigned integer value"},
+    {
+        "id": 5,
+        "type": "UInt16",
+        "description": "A 16 bit unsigned integer value",
+    },
     {"id": 6, "type": "Int32", "description": "A 32 bit signed integer value"},
-    {"id": 7, "type": "UInt32", "description": "A 32 bit unsigned integer value"},
+    {
+        "id": 7,
+        "type": "UInt32",
+        "description": "A 32 bit unsigned integer value",
+    },
     {"id": 8, "type": "Int64", "description": "A 64 bit signed integer value"},
-    {"id": 9, "type": "UInt64", "description": "A 64 bit unsigned integer value"},
+    {
+        "id": 9,
+        "type": "UInt64",
+        "description": "A 64 bit unsigned integer value",
+    },
     {
         "id": 10,
         "type": "Float",
@@ -679,9 +792,17 @@ TYPE_LIST = [
         "type": "Double",
         "description": "An IEEE double precision (64 bit) floating point value",
     },
-    {"id": 12, "type": "String", "description": "A sequence of Unicode characters"},
+    {
+        "id": 12,
+        "type": "String",
+        "description": "A sequence of Unicode characters",
+    },
     {"id": 13, "type": "DateTime", "description": "An instance in time"},
-    {"id": 14, "type": "Guid", "description": "A 128-bit globally unique identifier"},
+    {
+        "id": 14,
+        "type": "Guid",
+        "description": "A 128-bit globally unique identifier",
+    },
     {"id": 15, "type": "ByteString", "description": "A sequence of bytes"},
     {"id": 16, "type": "XmlElement", "description": "An XML element"},
     {
@@ -694,7 +815,11 @@ TYPE_LIST = [
         "type": "ExpandedNodeId",
         "description": "A node id that stores the namespace URI instead of the namespace index",
     },
-    {"id": 19, "type": "StatusCode", "description": "A structured result code"},
+    {
+        "id": 19,
+        "type": "StatusCode",
+        "description": "A structured result code",
+    },
     {
         "id": 20,
         "type": "QualifiedName",
@@ -715,10 +840,14 @@ TYPE_LIST = [
         "type": "DataValue",
         "description": "A data value with an associated quality and timestamp",
     },
-    {"id": 24, "type": "Variant", "description": "Any of the other built-in types"},
+    {
+        "id": 24,
+        "type": "Variant",
+        "description": "Any of the other built-in types",
+    },
     {
         "id": 25,
         "type": "DiagnosticInfo",
         "description": "A diagnostic information associated with a result code",
     },
-]
\ No newline at end of file
+]
